@@ -1,5 +1,7 @@
 import numpy as np
 from bice.core.equation import Equation
+from bice.core.types import Shape
+from typing import Optional
 
 
 class PartialDifferentialEquation(Equation):
@@ -10,7 +12,7 @@ class PartialDifferentialEquation(Equation):
     FiniteDifferencesEquation.
     """
 
-    def __init__(self, shape=None):
+    def __init__(self, shape: Optional[Shape] = None) -> None:
         super().__init__(shape)
         # the spatial coordinates
         if len(self.shape) > 0:
@@ -19,13 +21,14 @@ class PartialDifferentialEquation(Equation):
             self.x = None
 
     @property
-    def spatial_dimension(self):
+    def spatial_dimension(self) -> int:
         """Returns the spatial dimension of the domain self.x"""
+        assert self.x is not None
         if isinstance(self.x, np.ndarray):
             return self.x.ndim
         return len(self.x)
 
-    def du_dt(self, u=None):
+    def du_dt(self, u: Optional[np.ndarray] = None) -> np.ndarray:
         """calculate the time derivative du/dt of the unknowns"""
         # if u is not given, use self.u
         if u is None:
@@ -33,7 +36,7 @@ class PartialDifferentialEquation(Equation):
         # typically, the mass matrix determines which part of rhs(u) go into du/dt
         return self.mass_matrix().dot(self.rhs(u))
 
-    def du_dx(self, u=None, direction=0):
+    def du_dx(self, u: Optional[np.ndarray] = None, direction: int = 0) -> np.ndarray:
         """"
         Calculate the spatial derivative du/dx in a given spatial direction.
         (abstract, needs to be specified for child classes)
